/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */
#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <X11/cursorfont.h>
#include <Xm/DialogS.h>
#include <Xm/DrawingA.h>
#include <Xm/PushB.h>
#include <Xm/MessageB.h>
#include <Xm/Label.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/Frame.h>
#include <Xm/BulletinB.h>
#include <Xm/DrawnB.h>
#include <Xm/ScrollBar.h>
#include <Xm/Form.h>
#include <Xm/MainW.h>
#include <Xm/MwmUtil.h>
#include <X11/xpm.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include "xmap.h"
#include "mode.h"
#include "help/HTML2.h"
#include "debug.h"

extern XtAppContext app;
extern Widget toplevel;
extern XtTranslations hlpbutton_translations;

extern char *get_hlptip(Widget w);

static int bs;
static XtIntervalId id;
static XtWorkProcId work_id;

static Widget dialog, rc, toolbar, fr, help;
static void tool_cb();
static void unshow_tip();

static Cursor gotocursor;

#define HISTORY 64
static int hist[HISTORY];
static int history;

#define WIDTH 28

void activate(Widget w, XtPointer call_data, XfwfFTextCallbackStruct * cd);

enum {
    WAIT, DISPLAY, IGNORE
};

struct header {
    char header[64];
    long npages;
    long nitems;
    long offset;
};

static char *hf = NULL;
static int npages;
static int nitems;
static int current_page;
static int offset;
static char *page;
static int pagelen;

typedef struct {
    char *name;
    long length;
    long offset;
} Toc;

Toc *toc;



static String txt = "\
<BODY>\n\
<H1>Sorry!</H1>\n\
\n\
<P>No help on this topic.\n\
\n\
\n\
<\\BODY>\
";



static void push(int i)
{
    if (history<(HISTORY-1)) {
	hist[history] = i;
	history++;
    }
    else {
	memcpy(hist,hist+sizeof(int),(HISTORY-1)*sizeof(int));
	hist[history] = i;
    }
}

static int pop()
{
    if (history) history--;
    return hist[history];
}

Widget showHelpToolBarButton(int button, int ord, Pixmap * pmap, void *p, XtPointer p_data)
{
    Widget w;

    w = XtVaCreateWidget("tool", xmDrawnButtonWidgetClass, toolbar,
			 XmNlabelType, XmPIXMAP,
			 XmNlabelPixmap, pmap,
			 XmNhighlightThickness, 0,
			 XmNshadowThickness, 0,
			 XmNpushButtonEnabled, True,
			 XmNshadowType, XmSHADOW_ETCHED_IN,
		       XmNx, 6 + (ord * WIDTH) + tbutton[button].spacing,
			 XmNy, 3,
			 NULL);
    XtAddCallback(w, XmNactivateCallback, (XtCallbackProc) tool_cb, (void *) (button));
    if (p)
      XtAddCallback(w, XmNactivateCallback, (XtCallbackProc) p, (void *) p_data);
    hbutton[button].w = w;
    XtOverrideTranslations(w, hlpbutton_translations);
    XtManageChild(w);
    return w;
}

static void to_cb(Widget w)
{
    char bufp[256];
    id = 0;
    snprintf(bufp, 255, "%s", get_hlptip(w));
    do_tip(rc, w, bufp, 35, &options);
    bs = DISPLAY;
}

void HlpMotionAction(Widget w, XMotionEvent * e, String * p, Cardinal n)
{
    if (bs == WAIT) {
	if (id) XtRemoveTimeOut(id);
	id = XtAppAddTimeOut(app, 550L, (XtTimerCallbackProc) to_cb, w);
    }
#ifdef REMOVE_WIGGLE
    else if (bs == DISPLAY) {
	bs = IGNORE;
	unshow_tip();
    }
#endif
}

void HlpEnterAction(Widget w, XEvent * e, String * p, Cardinal n)
{
    bs = WAIT;
    XtVaSetValues(w, XmNshadowThickness, 1, NULL);
    if (id) XtRemoveTimeOut(id);
    id = XtAppAddTimeOut(app, 550L, (XtTimerCallbackProc) to_cb, w);
}

void HlpLeaveAction(Widget w, XEvent * e, String * p, Cardinal n)
{
    if (bs == WAIT && id) {
	XtRemoveTimeOut(id);
	id = 0;
    }
    XtVaSetValues(w, XmNshadowThickness, 0, NULL);
    if (bs == DISPLAY) {
	unshow_tip();
    }
}

static void unshow_tip()
{
    if (bs == WAIT && id) {
	XtRemoveTimeOut(id);
	id = 0;
    }
    if (bs == DISPLAY) {
	undo_tip(&options);
    }
    bs = IGNORE;
}

static void tool_cb(Widget w, XtPointer client_data, XmToggleButtonCallbackStruct * state)
{
    if (options.toolhelp)
	unshow_tip();

    /*
    if (hbutton[(int) client_data].func)
	hbutton[(int) client_data].func(w, NULL);
    */
}

static int init_help()
{
    struct header *p;
    char *cp;
    int i;
    long loff = 0;

    p = (struct header *) hf;

    if (strncmp("hlp07", hf, 5)) {
	error("Invalid Help File!\nNo Help available.",0);
	error_wait_ack();
	return 0;
    }

    npages = ntohl(p->npages);
    nitems = ntohl(p->nitems);
    offset = ntohl(p->offset);

    toc = (Toc *) calloc(nitems, sizeof(Toc));

    cp = hf + sizeof(struct header);

    for (i = 0; i < nitems; i++) {
	toc[i].name = cp;
	toc[i].offset = loff + offset;;
	toc[i].length = ntohl(MY_ALIGN((long *) ((cp + *cp) + 1L)));
	loff += toc[i].length;
	cp += *cp + sizeof(char) + sizeof(long);
    }
    return 1;
}

static char *get_img(char *url, int *length)
{
    int i;

    for (i = npages; i < nitems; i++) {
	if (strncmp(url, toc[i].name + 1, (int) *toc[i].name) == 0) {
	    *length = toc[i].length;
	    return hf + toc[i].offset;
	}
    }
    *length = 0;
    return NULL;
}

static char *get_page(char *url, int *length)
{
    int i;

    for (i = 0; i < npages; i++) {
	if (strncmp(url, toc[i].name + 1, (int) *toc[i].name) == 0) {
	    push(current_page);
	    current_page = i;
	    *length = toc[i].length;
	    return hf + toc[i].offset;
	}
    }
    beep();
    return NULL;
}

void resolve(Widget w, XtPointer call_data, XfwfResolveRec * cd)
{
    Pixmap pmap;
    char *p;
    XpmAttributes attr;
    Display *dpy = XtDisplayOfObject(toplevel);
    Screen *scr = XtScreenOfObject(toplevel);

    p = get_img(cd->url, &pagelen);
    attr.visual = DefaultVisual(dpy, DefaultScreen(dpy));
    attr.colormap = DefaultColormap(dpy, DefaultScreen(dpy));
    attr.depth = DefaultDepthOfScreen(scr);
    attr.exactColors = False;
    attr.closeness = 40000;

    attr.valuemask = XpmExactColors | XpmCloseness;

    if (XpmCreatePixmapFromBuffer(XtDisplay(toplevel),
             XtWindow(toplevel), p,
             &pmap, NULL,
             &attr) != XpmSuccess)
        fprintf(stderr, "Error reading help pixmap %s!\n", cd->url);

    XtVaSetValues(cd->widget,
	XtNwidth, attr.width,
	XtNheight, attr.height,
	XtNbackgroundPixmap, pmap,
	NULL);
}


static int new_page()
{
    if (help) {
    XtUnmanageChild(help);
    XtRemoveAllCallbacks(help,XtNactivate);
    XtRemoveAllCallbacks(help,XtNresolveURL);

//FIXME for now we are leaking memory here.
//I do not understand why we get a core when help is destroyed.
//    XtDestroyWidget(help); 
    }

    help = XtVaCreateManagedWidget("Help", xfwfHTML2WidgetClass, fr,
                                   XtNactiveCursor, gotocursor,
                                   NULL);
    XtAddCallback(help, XtNactivate, (XtCallbackProc) activate, NULL);
    XtAddCallback(help, XtNresolveURL, (XtCallbackProc) resolve, NULL);
    XfwfAddText(help, page, pagelen);
    XfwfAddText(help, NULL, 0);


    return True;
}

void activate(Widget w, XtPointer call_data, XfwfFTextCallbackStruct * cd)
{
    XfwfSSGMLData *pd = cd->data;

    page = get_page(pd->url, &pagelen);

    if (page) work_id = XtAppAddWorkProc(app, (XtWorkProc)new_page, NULL);
}

void create_help_dialog()
{
    Arg args[5];
    int n;
    int fd;
    size_t size;
    struct stat statbuf;
    char *home = getenv("HOME");
    char bufp[256];

#if 0
    dialog = XtVaCreatePopupShell("help",
				xmDialogShellWidgetClass, toplevel,
				XmNtitle, "Hugo Help",
				XmNdeleteResponse, XmUNMAP,
				XmNmarginWidth, 5,
				XmNmarginHeight, 5,
				NULL);
#else
    n = 0;
    XtSetArg(args[n], XmNtitle, "Hugo Help"); n++;
    XtSetArg(args[n], XmNmarginWidth, 5); n++;
    XtSetArg(args[n], XmNmarginHeight, 5); n++;

    dialog = XmCreateDialogShell(toplevel, "help", args, n);
#endif
    rc = XtVaCreateWidget("work_area", xmFormWidgetClass, dialog, NULL);

    toolbar = XtVaCreateWidget("toolbar", xmBulletinBoardWidgetClass, rc,
				XmNmarginHeight, 4,
				XmNmarginWidth, 4,
				XmNtopAttachment, XmATTACH_FORM,
				XmNleftAttachment, XmATTACH_FORM,
				XmNrightAttachment, XmATTACH_FORM,
				XmNheight, 34,
				XmNminHeight, 34,
				XmNmaxHeight, 34,
				XmNresizePolicy, XmRESIZE_NONE,
				NULL);

    fr = XtVaCreateWidget("fr1", xmFrameWidgetClass, rc,
				XmNleftAttachment, XmATTACH_FORM,
				XmNrightAttachment, XmATTACH_FORM,
				XmNbottomAttachment, XmATTACH_FORM,
				XmNtopAttachment, XmATTACH_WIDGET,
				XmNtopWidget, toolbar,
				NULL);

    gotocursor = XCreateFontCursor(XtDisplay(toplevel), XC_hand2);

    help = XtVaCreateWidget("Help", xfwfHTML2WidgetClass, fr,
				XtNactiveCursor, gotocursor,
				NULL);

    XtAddCallback(help, XtNactivate, (XtCallbackProc) activate, NULL);
    XtAddCallback(help, XtNresolveURL, (XtCallbackProc) resolve, NULL);


    if (!home) home = "./";
    snprintf(bufp, 255, "%s/hugo.hlp", home);

    if ((fd = open(bufp, O_RDONLY)) == -1) {
	snprintf(bufp, 255, "%s/hugo.hlp", COMMON_PATH );
	if ((fd = open(bufp, O_RDONLY)) == -1) {
	    error("Help file not found, no help available!\n",errno);
	    error_wait_ack();
	    return;
	}
    }
    fstat(fd, &statbuf);
    size = statbuf.st_size;
    hf = mmap(0, size, PROT_READ, MAP_SHARED, fd, 0);
    if (!init_help()) {
	close(fd);
	hf = NULL;
    }
}

void show_help(char *chapter)
{
    static sized = 0;
    int sx, sy, sw, sh;
    Position x, y;
    char bufp[256];

    if (hf) {
	if (!sized) {
	    sized++;
	    XtVaGetValues(toplevel,
			  XmNx, &x,
			  XmNy, &y,
			  NULL);

	    y += 50;
	    snprintf(bufp, 255, "%dx%d+%d+%d", 310, 500, x, y);
	    GetPrivateProfileString("Layout", "help", bufp, bufp, 255, 
				    inifile);

	    sscanf(bufp, "%dx%d+%d+%d", &sh, &sw, &sx, &sy);
	    XtVaSetValues(rc,
			  XmNx, sx,
			  XmNy, sy,
			  XmNwidth, sh,
			  XmNheight, sw,
			  NULL);
	}

	XtManageChild(rc);
	XtManageChild(toolbar);

	XtManageChild(dialog);

	if ( (page = get_page(chapter, &pagelen)) ) 
		;
	else if ( (page = get_page("index.html", &pagelen)) )
		;
	else {
	    page = txt;
	    pagelen = strlen(txt);
	}
	new_page();
	XtManageChild(fr);
	XtManageChild(help);
	XtPopup(dialog, XtGrabNone);
    }
    else beep();
}

void do_help()
{
	show_help("index.html");
}

void do_relnotes()
{
	show_help("RelNotes.html");
}

void do_credits()
{
	show_help("Credits.html");
}

void show_help_cb(Widget dialog, XtPointer client_data, XtPointer call_data)
{
	show_help( (char *)client_data);
}

void do_hexit()
{
	XtPopdown(dialog);
}

void save_help_properties()
{
    Position x, y;
    Dimension width, height;
    char bufp[256];

    if (dialog) {
	XtVaGetValues(rc,
		      XmNx, &x,
		      XmNy, &y,
		      XmNwidth, &width,
		      XmNheight, &height,
		      NULL);

	snprintf(bufp, 255, "%dx%d+%d+%d", width, height, x, y);
	if (width > 100 && height > 100 && x > 0 && y > 0)
	    WritePrivateProfileString("Layout", "help", bufp, inifile);
    }
}

void do_back()
{
    int ret;

    if (current_page != (ret = pop())) {
	current_page=ret;
	pagelen = toc[current_page].length;
	page = hf + toc[current_page].offset;
	work_id = XtAppAddWorkProc(app, (XtWorkProc)new_page, NULL);
    } else beep();
}

void do_next()
{
    if (current_page < (npages-1)) {
	push(current_page);
	current_page++;
	pagelen = toc[current_page].length;
	page = hf + toc[current_page].offset;
	work_id = XtAppAddWorkProc(app, (XtWorkProc)new_page, NULL);
    } else beep();
}

void do_previous()
{
    if (current_page > 0) {
	push(current_page);
	current_page--;
	pagelen = toc[current_page].length;
	page = hf + toc[current_page].offset;
	work_id = XtAppAddWorkProc(app, (XtWorkProc)new_page, NULL);
    } else beep();
}


#if 0
char *bad = "
/* XPM */\n
static char *bad[] = {\n
/* width height num_colors chars_per_pixel */\n
\"    22    25        8            1\",\n
/* colors */\n
\". c #000000\",\n
\"# c #008000\",\n
\"a c #303060\",\n
\"b c #808080\",\n
\"c c #c0c0c0\",\n
\"d c #f86430\",\n
\"e c #f864c8\",\n
\"f c #f8fcf8\",\n
/* pixels */\n
\"ccccccc#####ccccbfffcb\",\n
\"cccccc#######cccbffffc\",\n
\"ccccc#########ccbfffff\",\n
\"ccccc#########cc......\",\n
\"ccccc#########cccaaaaa\",\n
\"ccccc#########ccccbbbb\",\n
\"ccccc#########cccccccf\",\n
\"cccccc#######ccccccccf\",\n
\"ecccccc#####cccccccccf\",\n
\"eecccccccccccccccccccf\",\n
\"eeeccccccccccccccccccf\",\n
\"eeeecccccccccccccccccf\",\n
\"eeeeeccccccddddddddcca\",\n
\"eeeeeecccccddddddcccaa\",\n
\"eeeeeeeccccdddddccaaab\",\n
\"eeeeeecccccddcccaaabbc\",\n
\"eeeeeccccccccccaabbccc\",\n
\"eeeeccccccaaaa.abccccc\",\n
\"eeeccccccaabbbbccccccc\",\n
\"eecccccccabccccccccccb\",\n
\"eccccaaaaabcccccccccbf\",\n
\"cccccabbbcccccccccbbcf\",\n
\"ccccaabccccccccbcccccf\",\n
\"ffaabbccccccccbcffffff\",\n
\"faabccccccccccffffffff\"\n
};\n
";
#endif
