/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */

#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>

#include <X11/Xos.h>
#include <Xm/DrawingA.h>

#include "xmap.h"
#include "mode.h"
#include "debug.h"
#include "print.h"

extern Widget draw_w;

FILE *ps;
int medium = SCREEN;

double paperheight = PAPER_LETTER_HEIGHT;
double paperwidth = PAPER_LETTER_WIDTH;
Dimension pixmap_height;
Dimension pixmap_width;

static unsigned long current_color;
static Page page;

int intensity_map[4096];	/* max size of color map is 4096 (I hope) */

XImage *image;

/* this is a structure containg the values of all command line options
   initially, we assume no options are set */
Flag flag =
{FALSE, FALSE, FALSE, FALSE, FALSE,
 FALSE, FALSE, FALSE, FALSE, FALSE,
 FALSE, FALSE, FALSE, FALSE, FALSE};

char *ColorImage[] =
{
    "systemdict /colorimage known not {        % only create if not in systemdict",
    "  % Utility procedure for colorimage operator.  This procedure takes a",
    "  % string of rgb encoded values and creates a string 1/3 as long with",
    "  % monochrome values.  This procedure assumes 8 bits/color (i.e. ",
    "  % 1 character/color)",
    "  % storage format for input string:  (r1 g1 b1  r2 g2 b2  r3 g3 b3  ... )",
    "  % storage format for output string: (g1  g2  g3 ... )",
    "  ",
    "  /colortograyscale { %def                % (string)",
    "    dup /rgbdata exch store               % (string)",
    "    length 3 idiv                         % Ns/3 ",
    "    /npixls exch store                    % ; npixls => Ns/3",
    "    /indx 0 store                         % ; indx => 0",
    "    /pixls npixls string store            % ; pixls => (....)",
    "    0 1 npixls -1 add {                   % counter ",
    "      pixls exch                          % pixls counter",
    "      rgbdata indx get .3 mul             % pixls counter .3*rgbdata(ind)",
    "      rgbdata indx 1 add get .59 mul add  % pixls counter .3*rgbdata(ind) + ",
    "                                          %          .59*rgbdata(ind+1)",
    "      rgbdata indx 2 add get .11 mul add  % pixls counter .3*rgbdata(ind) + .59",
    "                                          %  *rgbdata(ind+1)+.11*rgbdata(ind+2)",
    "      cvi                                 % pixls counter <grayscale value>",
    "      put                                 %",
    "      /indx indx 3 add store              % ; /ind => ind+3",
    "    } for                                 % repeat for each rgb value",
    "    pixls                                 % (pixls)",
    "  } bind def                              % ; /colortograyscale -> dictionary",
    "  ",
    "  % Utility procedure for colorimage operator.  This procedure takes two",
    "  % procedures off the stack and merges them into a single procedure.",
    "  ",
    "  /mergeprocs { %def      % {proc1} {proc2}",
    "    dup length            % {proc1} {proc2} N2",
    "    3 -1 roll             % {proc2} N2 {proc1}",
    "    dup                   % {proc2} N2 {proc1} {proc1}",
    "    length                % {proc2} N2 {proc1} N1",
    "    dup                   % {proc2} N2 {proc1} N1 N1",
    "    5 1 roll              % N1 {proc2} N2 {proc1} N1",
    "    3 -1 roll             % N1 {proc2} {proc1} N1 N2",
    "    add                   % N1 {proc2} {proc1} N1+N2",
    "    array cvx             % N1 {proc2} {proc1} { ... }",
    "    dup                   % N1 {proc2} {proc1} { ... } { ... }",
    "    3 -1 roll             % N1 {proc2} { ... } { ... } {proc1}",
    "    0 exch                % N1 {proc2} { ... } { ... } 0 {proc1}",
    "    putinterval           % N1 {proc2} { <<{proc1}>> ... }",
    "    dup                   % N1 {proc2} { <<{proc1}>> ... } { <<{proc1}>> ... }",
    "    4 2 roll              % { <<{proc1}>> ... } { <<{proc1}>> ... } N1 {proc2}",
    "    putinterval           % { <<{proc1}>> <<{proc2}>> }",
    "  } bind def              % ; /mergeprocs => dictionary",
    "",
    "  /colorimage { %def               % {imageproc} multiproc ncolors",
    "     pop                           % {imageproc} multiproc ; assume 3 colors",
    "     pop                           % {imageproc}           ; assume false",
    "     {colortograyscale}            % {imageproc} {colortograyscale}",
    "     mergeprocs                    % {imageproc colortograyscale}",
    "     image                         % construct monochrome image",
    "  } bind def                       % ; /colorimage => dictionary",
    "} if                               % only create if it doesn't already exist",
    0
};

char *ErrorHandler[] =
{
    "",
    "errordict begin",
    "  /handleerror {",
    "  systemdict begin",
    "    initgraphics",
    "    /Courier findfont 12 scalefont setfont",
    "    72 720 moveto",
    "    (Postscript Problem: ) show",
    "    $error /errorname get",
    "    =string cvs show",
    "    72 700 moveto",
    "    (Command: ) show",
    "    $error /command get",
    "    =string cvs show",
    "    systemdict /showpage get exec",
    "    end",
    "  } def",
    "end",
    0
};


outputImage(char **image)
{
    while (*image)
	fprintf(ps, "%s\n", *image++);
}

void set_ps_color(GC gc)
{
    XGCValues gcv;

    XGetGCValues(XtDisplay(draw_w), gc, GCForeground, &gcv);
    if (gcv.foreground != current_color) {
	XColor color;
	Display *dpy = XtDisplay(draw_w);
	Colormap cmap = DefaultColormapOfScreen(XtScreen(draw_w));

	current_color = gcv.foreground;

	color.pixel = current_color;
	XQueryColor(dpy, cmap, &color);

	fprintf(ps, "%.3f %.3f %.3f setrgbcolor\n",
		(double) color.red / (double) 0xffff,
		(double) color.green / (double) 0xffff,
		(double) color.blue / (double) 0xffff);
    }
}

get_page_info(Page * the_page)
{
    /* letter paper size 8.5 x 11 */
    switch (the_page->type) {
    case LETTER:
	the_page->xcenter = 8.5 / 2;
	the_page->ycenter = 11.0 / 2 + .21;
	the_page->maxwsize = 8.11;
	the_page->maxhsize = 8.91;
	the_page->default_width = 7.0;
	the_page->default_height = 8.5;
	return (0);

    case LEGAL:
	the_page->xcenter = 11.0 / 2;
	the_page->ycenter = 17.0 / 2 + .21;
	the_page->maxwsize = 10.60;
	the_page->maxhsize = 14.91;
	the_page->default_width = 9.0;
	the_page->default_height = 14.5;
	return (0);

    case A4:
	the_page->xcenter = 8.3 / 2;
	the_page->ycenter = 11.7 / 2 + .21;
	the_page->maxwsize = 8.26;
	the_page->maxhsize = 9.60;
	the_page->default_width = 7.25;
	the_page->default_height = 9.0;
	return (0);

    case A3:
	the_page->xcenter = 11.7 / 2;
	the_page->ycenter = 16.6 / 2 + .21;
	the_page->maxwsize = 11.69;
	the_page->maxhsize = 16.53;
	the_page->default_width = 9.0;
	the_page->default_height = 14.0;
	return (0);

    default:
	return (1);
    }
}


printEPSF(Image image)
{
    long clock;
    struct passwd *pswd;	/* variables for userid, date and time */
    char hostname[256];		/* name of host machine */

#if EPS
    fprintf(ps, "%%!PS-Adobe-2.0 EPSF-2.0\n");	/* standard PS header */

    if (image.orientation == PORTRAIT)
	fprintf(ps, "%%%%BoundingBox: %f %f %f %f\n",
		72 * page.ximagepos - 1, 72 * page.yimagepos - 1,
		72 * (page.ximagepos + image.width) + 1,
		72 * (page.yimagepos + image.height - page.yoffset) + 1);
    else
	fprintf(ps, "%%%%BoundingBox: %f %f %f %f\n",
		72 * page.ximagepos - 1, 72 * page.yimagepos - 1,
		72 * (page.ximagepos + image.height) + 1,
		72 * (page.yimagepos + image.width - page.yoffset) + 1);
#else
    fprintf(ps, "%%!PS-Adobe-2.0\n");	/* standard PS header */
#endif

    fprintf(ps, "%%%%Pages: 1\n");
    fprintf(ps, "%%%%Creator: Hugo map mover\n");
    fprintf(ps, "%%%%CreationDate: %s", (time(&clock), ctime(&clock)));
    fprintf(ps, "%%%%Title: %s\n", mapdata.map->description);

    pswd = getpwuid(getuid());
    gethostname(hostname, sizeof hostname);
    fprintf(ps, "%%%%For: %s:%s (%s)\n", hostname, pswd->pw_name, pswd->pw_gecos);
    fprintf(ps, "%%%%EndComments\n");


    fprintf(ps, "%%%%BeginProlog\n\n");
    fprintf(ps, "/inch {72 mul} def\n");
    fprintf(ps, "%%%%EndProlog\n\n");
    fprintf(ps, "%%%%Page: 1 1\n\n");
    fprintf(ps, "%%%%BeginPageSetup\n");
    if (options.pdebug) outputImage(ErrorHandler);
    fprintf(ps, "gsave\n");
    fprintf(ps, "%%%%EndPageSetup\n");
}


print_map()
{
    char temp[128];
    int dest_is_file;

    XtVaGetValues(draw_w,
		  XmNwidth, &pixmap_width,
		  XmNheight, &pixmap_height,
		  NULL);

    GetPrivateProfileString("Printing", "destination", "file", temp, 128, inifile);
    dest_is_file = strcmp(temp,"printer");

    if (dest_is_file) {
	GetPrivateProfileString("Printing", "filename", "xmap.ps", temp, 128, inifile);
	ps = fopen(temp, "w");
    }
    else {
	GetPrivateProfileString("Printing", "printcommand", "lpr", temp, 128, inifile);
	ps = popen(temp, "w");
    }

    if (!ps) error("Can't print!",errno);
    else {
	print_image();
	if (dest_is_file) fclose(ps);
	else pclose(ps);
    }
}

print_vector()
{
    int mapWidth, mapHeight;
    double mapWidthcm, mapHeightcm;
    double maxhcheck, maxwcheck;

    medium = COLORPS;

    fprintf(ps, "newpath\n");
    fprintf(ps, "0 0 moveto\n");
    fprintf(ps, "0 %d lineto\n", pixmap_height);
    fprintf(ps, "%d %d lineto\n", pixmap_width, pixmap_height);
    fprintf(ps, "%d 0 lineto\n", pixmap_width);
    fprintf(ps, "0 0 lineto\n");
    fprintf(ps, "closepath\n");
    fprintf(ps, "clip\n");

    vprint();

    fprintf(ps, "newpath\n");
    fprintf(ps, "0 0 moveto\n");
    fprintf(ps, "0 %d lineto\n", pixmap_height);
    fprintf(ps, "%d %d lineto\n", pixmap_width, pixmap_height);
    fprintf(ps, "%d 0 lineto\n", pixmap_width);
    fprintf(ps, "0 0 lineto\n");
    fprintf(ps, "closepath\n");
    fprintf(ps, "stroke\n");

    medium = SCREEN;
}



print_image()
{
    Display *dpy = XtDisplay(draw_w);
    XColor  xcolor, ncolor;
    Colormap cmap = DefaultColormapOfScreen(XtScreen(draw_w));

    /* variables for image size */
    Image my_image;

    /* variables for image data */
    unsigned char intens;
    int color, runlen, gray_index, intensity;
    int c, i, j, im, ig;
    int outputcount;
    int line_skip, line_end, col_skip, col_start, col_end;
    long swaptest = 1;

    /*  page size and image positioning */
    Page page;

    float maxhcheck, maxwcheck;

    float ha, wa;

    unsigned char rr, gg, bb;
    char s_translate[128];	/* PostScript code to position image */
    char s_scale[80];		/* PostScript code to scale image */
    char s_matrix[80];		/* PostScript code for image matrix */

    /* initializations */
    page.type = GetPrivateProfileInt("Printing", "papersize", 0, inifile);

    page.yoffset = 0.0;		/* position adjustment */
    page.height_adjust = 0.0;	/* image height adjustment */
    page.time_date_height = 0.0;	/* height (inches) of time & date characters   */
    my_image.height_frac = 1.0;
    GetPrivateProfileString("Printing", "orientation", "portrait", s_translate, 128, inifile);
    my_image.orientation = strcmp(s_translate,"portrait")?LANDSCAPE:PORTRAIT;
 


    /* Get page size information */
    get_page_info(&page);


    my_image.ps_width = pixmap_width;
    col_skip = 0;

    col_start = 1;
    col_end = col_skip + my_image.ps_width;

    my_image.ps_height = pixmap_height;
    line_skip = 0;

    line_end = my_image.ps_height;

    my_image.pixels_width = my_image.ps_width;	/* TEMP - this might have to be fixed later */

#if 0
    if (flag.w == FALSE && flag.h == TRUE)
	my_image.width = (my_image.height * my_image.pixels_width) / my_image.ps_height;

    if (flag.w == TRUE && flag.h == FALSE)
	my_image.height = (my_image.width * my_image.ps_height) / my_image.pixels_width;
#endif

#if 1
    /*
     * choose defaults to make image as large as possible on page
     */
    if (flag.w == FALSE && flag.h == FALSE) {
	/* setup page sizes for orientation requested */
	if (my_image.orientation == PORTRAIT) {
	    wa = page.default_width;
	    ha = page.default_height - page.height_adjust;
	} else {
	    wa = page.default_height;
	    ha = page.default_width - page.height_adjust;
	}
	if (wa * my_image.ps_height >= ha * my_image.pixels_width) {
	    my_image.height = ha;
	    my_image.width = (my_image.height * my_image.pixels_width) / my_image.ps_height;
	} else {
	    my_image.width = wa;
	    my_image.height = (my_image.width * my_image.ps_height) / my_image.pixels_width;
	}
    }
#endif

    /* 
     * Center & scale image and setup image matrix
     */
    if (my_image.orientation == PORTRAIT) {
	maxwcheck = page.maxwsize;
	maxhcheck = page.maxhsize;
	page.ximagepos = page.xcenter - (my_image.width / 2);
	page.yimagepos = page.ycenter - (my_image.height / 2) + page.yoffset;
	sprintf(s_matrix, "  [%d 0 0 -%d 0 %d]",
	  my_image.pixels_width, my_image.ps_height, my_image.ps_height);
    } else {
	maxwcheck = page.maxhsize;
	maxhcheck = page.maxwsize;
	page.ximagepos = page.xcenter - (my_image.height / 2);
	page.yimagepos = page.ycenter - (my_image.width / 2) + page.yoffset;
	sprintf(s_matrix, "  [0 %d %d 0 0 0]", my_image.pixels_width, my_image.ps_height);
    }
    sprintf(s_translate, "%f inch %f inch translate", page.ximagepos, page.yimagepos);
    sprintf(s_scale, "%f inch %f inch scale", my_image.width, my_image.height);

    set_cursor(True);

    printEPSF(my_image);

    if (flag.gamma == TRUE)
	fprintf(ps, "{ %f exp } settransfer\n", my_image.gamma);

    if (mapdata.type == VECTOR) {
	fprintf(ps, "0.4 inch %.1f inch translate\n", maxhcheck);
	fprintf(ps, "0.5 0.5 scale\n");
	fprintf(ps, "1 -1 scale\n");
	print_vector();
    } else {
	if (flag.mono == FALSE) {	/* print as color image */
	    fprintf(ps, "/buffer 4 string def\n");
	    fprintf(ps, "/rgb 3 string def\n");
	    outputImage(ColorImage);
	    fprintf(ps, "/drawfullcolorimage {\n");
	    fprintf(ps, "  %4d %4d 8\n", my_image.pixels_width, my_image.ps_height);
	    fprintf(ps, "%s\n", s_matrix);
	    fprintf(ps, "  { currentfile buffer readhexstring pop pop %% get runlength and rgb values\n");
	    fprintf(ps, "    /npixels buffer 0 get 1 add 3 mul store  %% number of pixels\n");
	    fprintf(ps, "    /pixels npixels string store             %% create string to hold rgb values\n");
	    fprintf(ps, "    /rgb buffer 1 3 getinterval store        %% red green blue values\n");
	    fprintf(ps, "    0 3 npixels -1 add {\n");
	    fprintf(ps, "      pixels exch rgb putinterval\n");
	    fprintf(ps, "    } for\n");
	    fprintf(ps, "  pixels                                     %% return string of colors\n");
	    fprintf(ps, "  }\n");
	    fprintf(ps, "  false 3\n");
	    fprintf(ps, "  colorimage\n");
	    fprintf(ps, "} bind def\n");
	    fprintf(ps, "\n");
	    fprintf(ps, "%s\nmatrix currentmatrix\n%s\n", s_translate, s_scale);
	    fprintf(ps, "\ndrawfullcolorimage\n");
	} else {		/*  print as monochrome image */
	    fprintf(ps, "/buffer 2 string def\n");
	    fprintf(ps, "/drawimage {\n");
	    fprintf(ps, "  %4d %4d 8\n", my_image.pixels_width, my_image.ps_height);
	    fprintf(ps, "%s\n", s_matrix);
	    fprintf(ps, "  { currentfile buffer readhexstring pop pop\n");
	    fprintf(ps, "    /npixels buffer 0 get store\n");
	    fprintf(ps, "    /grayscale buffer 1 get store\n");
	    fprintf(ps, "    /pixels npixels 1 add string store\n");
	    fprintf(ps, "       grayscale 0 ne {\n");
	    fprintf(ps, "         0 1 npixels {pixels exch grayscale put } for\n");
	    fprintf(ps, "       } if\n");
	    fprintf(ps, "    pixels\n");
	    fprintf(ps, "  }\n");
	    fprintf(ps, "  image\n");
	    fprintf(ps, "} bind def\n");
	    fprintf(ps, "\n");
	    fprintf(ps, "%s\nmatrix currentmatrix\n%s\n", s_translate, s_scale);
	    fprintf(ps, "\ndrawimage\n");
	}

	/* start outputting the image */
	outputcount = 0;
	image = XGetImage(dpy, mapdata.map->pmap, 0, 0, pixmap_width, pixmap_height,
		XAllPlanes(), ZPixmap);


	for (i = 0; i < my_image.ps_height; i++) {
	    xcolor.pixel = XGetPixel(image, col_skip, i );
	    runlen = 0;

	    for (j = col_start; j < col_end; j++) {
		/*
		 * If the colors are the same, just keep counting, but make sure that
		 * the run length will fit in a byte (256 max, i.e. 0 means length 1,
		 * 1 means length 2, etc.)
		 */
		ncolor.pixel = XGetPixel(image, j, i );
		if ( xcolor.pixel == ncolor.pixel ) {
		    if (runlen == 0xff) {
			XQueryColor(dpy, cmap, &xcolor);
			rr = xcolor.red/256;
			gg = xcolor.green/256;
			bb = xcolor.blue/256;
			if (flag.mono == FALSE) {
			    fprintf(ps, "%02x%02x%02x%02x", runlen, rr, gg, bb);
			    outputcount += 4;
			} else {
			    intens = 0.299 * rr + 0.587 * gg + 0.114 * bb;
			    fprintf(ps, "%02x%02x", runlen, intens);
			    outputcount += 2;
			}
			runlen = -1;
		    }
		    runlen++;
		}
		/* if the color changed, then output the last run and reset counter */
		else {
		    XQueryColor(dpy, cmap, &xcolor);
		    rr = xcolor.red/256;
		    gg = xcolor.green/256;
		    bb = xcolor.blue/256;
		    if (flag.mono == FALSE) {
			fprintf(ps, "%02x%02x%02x%02x", runlen, rr, gg, bb);
			outputcount += 4;
		    } else {
			intens = 0.299 * rr + 0.587 * gg + 0.114 * bb;
			fprintf(ps, "%02x%02x", runlen, intens);
			outputcount += 2;
		    }
		    runlen = 0;
		    xcolor.pixel = ncolor.pixel;
		}
		/* check to make sure the output lines are not too long */
		if (outputcount >= MAXPERLINE) {
		    outputcount = 0;
		    fprintf(ps, "\n");
		}
	    }
	    XQueryColor(dpy, cmap, &xcolor);
	    rr = xcolor.red/256;
	    gg = xcolor.green/256;
	    bb = xcolor.blue/256;
	    if (flag.mono == FALSE) {
		fprintf(ps, "%02x%02x%02x%02x", runlen, rr, gg, bb);
		outputcount += 4;
	    } else {
		intens = 0.299 * rr + 0.587 * gg + 0.114 * bb;
		fprintf(ps, "%02x%02x", runlen, intens);
		outputcount += 2;
	    }
	}
	if (image) XDestroyImage(image);           /* clear old image out */
	image = NULL;

	fprintf(ps, "\n");
	fprintf(ps, "\nsetmatrix\n");
    }


    /* finish up PostScript code */
    fprintf(ps, "\n/#copies %d def\n", 1);
    fprintf(ps, "showpage grestore\n%%%%Trailer\n");
#if EPS
    fprintf(ps, "%%%%EOF\n");
#endif
    set_cursor(False);
}

